// Digital Cash
// Steven Kappes / Justin Fiedler

import java.util.*;
import java.math.*;
import java.security.*;
import javax.crypto.*;

public class Customer
{
	private int balance;
	public String ERROR = ""; 
	// Requesting from the server
	private boolean hasRequest; // If the customer is waiting on a withdrawl request
	private ArrayList<SecretKey> moEncryptKeys; // Keys the requested money orders are encrypted with
	private ArrayList<ByteArray> blindedHashes;
	private ArrayList<BigInteger> blindingFactors; // Blinds to the hash
	private ArrayList<MoneyOrder> pendingOrders;
	private int futureMOIndex; // Index of the money order the bank did not request to be unblinded

	// Digital cash
	private ArrayList<MoneyOrder> validMoneyOrders;

	public Customer()
	{
		balance = 0;
		hasRequest = false;

		validMoneyOrders = new ArrayList<MoneyOrder>();
	}

	private ArrayList<MoneyOrder> generateMoneyOrders(int amount)
	{
		ArrayList<MoneyOrder> orders = new ArrayList<MoneyOrder>();

		int indexToFake = (int)(DigitalCash.N * Math.random());

		System.out.println("Index to fake: " + indexToFake);
		for (int i = 0; i < DigitalCash.N; i++) {
			MoneyOrder order = new MoneyOrder((i == indexToFake) ? amount+1000000 : amount);
			orders.add(order);
		}

		return orders;
	}

	// Build N money orders which are blinded and signed
	public ArrayList<ByteArray> generateEncryptedMoneyOrders(int amount)
	{
		ArrayList<ByteArray> encryptOrders = new ArrayList<ByteArray>();
		int i;

		if (hasRequest) return encryptOrders; // Do not generate more orders if we are requesting one

		pendingOrders = this.generateMoneyOrders(amount);
		moEncryptKeys = new ArrayList<SecretKey>();
		blindingFactors = new ArrayList<BigInteger>();
		blindedHashes = new ArrayList<ByteArray>();

		for (i = 0; i < pendingOrders.size(); i++) {
			ByteArray mo = pendingOrders.get(i).intString();
			ByteArray moHash = pendingOrders.get(i).hash();
			SecretKey key = CryptoPrimitives.generateAESKey();
			BigInteger blindFactor = CryptoPrimitives.getNewBlindFactor(DigitalCash.getBankKey().getModulus());

			// Encrypt the order
			mo = CryptoPrimitives.AESencrypt(mo, key);

			// Apply the blinding factor to the hash
			moHash = new ByteArray(CryptoPrimitives.blind(new BigInteger(1, moHash.toByteArray()),
					blindFactor, DigitalCash.getBankKey()).toByteArray());

			encryptOrders.add(mo); // Save the encrypted money order
			blindedHashes.add(moHash); // Save the blinded hash
			moEncryptKeys.add(key);
			blindingFactors.add(blindFactor);
		}

		hasRequest = true;
		futureMOIndex = -1;
		return encryptOrders;
	}

	public ArrayList<ByteArray> getBlindedHashes()
	{
		if (!hasRequest) return null;
		return blindedHashes;
	}

	// Store the index of the money order that will be signed by the bank
	public void setFutureOrderIndex(int orderIndex)
	{
		futureMOIndex = orderIndex;
	}

	public int getFutureOrderIndex()
	{
		return futureMOIndex;
	}

	// Get all the encryption keys
	public ArrayList<SecretKey> getAllEncryptionKeys(int desiredMoneyOrderIndex)
	{
		ArrayList<SecretKey> keys = new ArrayList<SecretKey>();

		for (int i = 0; i < DigitalCash.N; i++) {
			if (i != desiredMoneyOrderIndex) {
				keys.add(this.getEncryptionKey(i));
			}
		}
		return keys;
	}

	// Return the encryption key for the specified money order
	public SecretKey getEncryptionKey(int desiredMoneyOrderIndex)
	{
		if (hasRequest) {
			if ((desiredMoneyOrderIndex < 0) && (desiredMoneyOrderIndex >= moEncryptKeys.size())) {
				System.err.println("Customer.getEncryptionKeys: Invalid index");
				System.exit(1);
			}
			return moEncryptKeys.get(desiredMoneyOrderIndex);
		} else {
			System.err.println("Customer.getEncryptionKey: No pending request");
			System.exit(1);	
		}
		return null;
	}

	// Get all the unblind factors
	public ArrayList<BigInteger> getAllBlinds(int desiredMoneyOrderIndex)
	{
		ArrayList<BigInteger> blinds = new ArrayList<BigInteger>();

		for (int i = 0; i < DigitalCash.N; i++) {
			if (i != desiredMoneyOrderIndex)
				blinds.add(this.getUnblindFactor(i));
		}

		return blinds;
	}

	// Return an unblind factor for the specified money order
	public BigInteger getUnblindFactor(int desiredMoneyOrderIndex)
	{
		if (hasRequest) {
			if ((desiredMoneyOrderIndex < 0) && (desiredMoneyOrderIndex >= blindingFactors.size())) {
				System.err.println("Customer.getUnblindFactor: Invalid index");
				System.exit(1);
			}
			return blindingFactors.get(desiredMoneyOrderIndex);
		} else {
			System.err.println("Customer.getUnblindFactor: No pending request");
			System.exit(1);	
		}
		return null;
	}

	// Get either the left or right secret keys for the identity strings
	public ArrayList<SecretKey> getIdentityKeys(int desiredMoneyOrderIndex, boolean leftKeys)
	{
		if (hasRequest) {
			if ((desiredMoneyOrderIndex < 0) && (desiredMoneyOrderIndex >= blindingFactors.size())) {
				System.err.println("Customer.getIdentityKeys: Invalid index");
				System.exit(1);
			}

			if (leftKeys)
				return pendingOrders.get(desiredMoneyOrderIndex).encryptKeysLeft();
			else
				return pendingOrders.get(desiredMoneyOrderIndex).encryptKeysRight();
		} else {
			System.err.println("Customer.getIdentityKeys: No pending request");
			System.exit(1);	
		}
		return null;
	}

	// Received a signed money order from the bank
	public boolean receivedSignedOrder(ByteArray signedHash)
	{
		if (hasRequest) {
			MoneyOrder mo;

			// Unblind the hash
			BigInteger unblinded = new BigInteger(1, signedHash.toByteArray());
			unblinded = CryptoPrimitives.unblind(unblinded, blindingFactors.get(futureMOIndex), DigitalCash.getBankKey());

			signedHash = new ByteArray(unblinded.toByteArray());
			mo = pendingOrders.get(futureMOIndex);
			mo.setSignedHash(signedHash);

			if (!mo.isValidOrder()) return false;

			this.addMoneyOrder(mo);
			hasRequest = false;
			return true;
		}
		return false;
	}

	// Add a money order signed by the bank
	public void addMoneyOrder(MoneyOrder mo)
	{
		validMoneyOrders.add(mo);
		balance -= mo.amount();
	}

	public ArrayList<SecretKey> getIdentKeys(ByteArray moID, ArrayList<Integer> unlock)
	{
		ArrayList<SecretKey> identKeys = new ArrayList<SecretKey>();
		MoneyOrder mo = this.getMoneyOrder(moID);

		for (int i = 0; i < unlock.size(); i++) {
			if (unlock.get(i).intValue() == 0)
				identKeys.add(mo.encryptKeysLeft().get(i));
			else
				identKeys.add(mo.encryptKeysRight().get(i));
		}

		return identKeys;			
	}

	// Block the customer if a request is pending
	public boolean isRequesting()
	{
		return hasRequest;
	}

	public void setHasRequest(boolean req)
	{
		hasRequest = req;
	}

	// Digital cash the customer currently has
	public ArrayList<MoneyOrder> validMoneyOrders()
	{
		return validMoneyOrders;
	}
	
	public int getBalance()
	{
		return balance;
	}

	private MoneyOrder getMoneyOrder(ByteArray moID)
	{
		for (int i = 0; i < validMoneyOrders.size(); i++) {
			if (moID.equals(validMoneyOrders.get(i).uniqueString()))
				return validMoneyOrders.get(i);
		}

		return null;
	}
}
