// Digital Cash
// Steven Kappes / Justin Fiedler

import java.util.*;
import java.math.*;
import java.security.*;
import javax.crypto.*;

public class MoneyOrder
{
	private ByteArray signedHash; // Stores the hash of the money order signed by the bank
	private ByteArray uniqueStr;
	private int amount;
	private boolean isSpent;
	private ArrayList<ByteArray> identityL;
	private ArrayList<ByteArray> identityR;
	private ArrayList<SecretKey> encryptL;
	private ArrayList<SecretKey> encryptR;
	private ArrayList<Integer> unlockIdent;

	// Only use for a placeholder money order
	public MoneyOrder()
	{
		amount = -1;		
	}

	// Create a new money order
	public MoneyOrder(int amt)
	{
		this.initVariables();

		// Build a money order
		uniqueStr = CryptoPrimitives.UniqueString(DigitalCash.UNIQUE_STR_SIZE);
		amount = amt;

		this.buildIdentityStrings();
	}

	// Store the signed money order
	// Decode the values from the unsigned money
	public MoneyOrder(ByteArray sData, ByteArray data)
	{
		this.initVariables();

		if (sData != null) signedHash = new ByteArray(sData.toInteger()); // Copy to integer first so the zeros are removed

		// Decode
		this.decodeDataString(data);
	}

	private void initVariables()
	{
		isSpent = false;
		identityL = new ArrayList<ByteArray>();
		identityR = new ArrayList<ByteArray>();
		encryptL = new ArrayList<SecretKey>();
		encryptR = new ArrayList<SecretKey>();
		unlockIdent = new ArrayList<Integer>();
	}

	public static boolean fake = false;

	private void buildIdentityStrings()
	{
		SecretKey key;
		ByteArray identity = DigitalCash.getMyIdentity();
		ByteArray secret1;
		ByteArray secret2;


		for (int i = 0; i < DigitalCash.N; i++) {
			// Secretly split the identity strings
			secret1 = CryptoPrimitives.UniqueString(DigitalCash.IDENTITY_STR_SIZE);
			if (fake) {
				secret2 = CryptoPrimitives.XOR(CryptoPrimitives.UniqueString(DigitalCash.IDENTITY_STR_SIZE), secret1);
			} else {
				secret2 = CryptoPrimitives.XOR(identity, secret1);
			}
			// Save the secrets with their hash value appended
			key = CryptoPrimitives.generateAESKey();
			identityL.add(CryptoPrimitives.AESencrypt(CryptoPrimitives.AppendHash(secret1), key));
			encryptL.add(key);

			key = CryptoPrimitives.generateAESKey();
			identityR.add(CryptoPrimitives.AESencrypt(CryptoPrimitives.AppendHash(secret2), key));
			encryptR.add(key);
		}

		// Check that the encoded identity strings are the correct length
		if (identityL.get(0).toByteArray().length != DigitalCash.IDENTITY_ENCRYPT_SIZE) {
			System.err.println("Exception (MoneyOrder.buildIdentityStrings): Invalid identity encryption size");
			System.exit(1);
		}
	}

	private void decodeDataString(ByteArray data)
	{
		byte [] dataStr = data.toByteArray();
		byte [] tempData;
		int index = 0;

		// Decode the unique string
		tempData = new byte[DigitalCash.UNIQUE_ENCRYPT_SIZE];
		DigitalCash.CopySomeBytes(tempData, dataStr, index);
		uniqueStr = new ByteArray(tempData);
		index += DigitalCash.UNIQUE_ENCRYPT_SIZE;

		// Decode the identity strings
		tempData = new byte[DigitalCash.IDENTITY_ENCRYPT_SIZE];
		for (int i = 0; i < DigitalCash.N; i++) {
			DigitalCash.CopySomeBytes(tempData, dataStr, index);
			identityL.add(new ByteArray(tempData));
			index += DigitalCash.IDENTITY_ENCRYPT_SIZE;

			DigitalCash.CopySomeBytes(tempData, dataStr, index);
			identityR.add(new ByteArray(tempData));
			index += DigitalCash.IDENTITY_ENCRYPT_SIZE;
		}

		// Decode the amount
		amount = DigitalCash.byteArrayToInt(dataStr, index);
	}

	public int amount()
	{
		return amount;
	}

	public ByteArray uniqueString()
	{
		return uniqueStr;
	}

	public ByteArray getSignedHash()
	{
		return signedHash;
	}

	public void setSignedHash(ByteArray signed)
	{
		signedHash = signed;
	}

	public ArrayList<ByteArray> identityStringsLeft()
	{
		return identityL;
	}

	public ArrayList<ByteArray> identityStringsRight()
	{
		return identityR;
	}

	public ArrayList<SecretKey> encryptKeysLeft()
	{
		return encryptL;
	}

	public ArrayList<SecretKey> encryptKeysRight()
	{
		return encryptR;
	}
	
	public ArrayList<Integer> unlockedIdentitySecrets()
	{
		return unlockIdent;
	}
	
	public void unlockIdentSecret(Integer val)
	{
		unlockIdent.add(val);
	}

	public void addSecretIdentKey(SecretKey key)
	{
		int index = encryptL.size();

		// Keep each list the same size
		if (unlockIdent.get(index).intValue() == 0) {
			encryptL.add(key);
			encryptR.add(DigitalCash.NULL_KEY);
		} else {
			encryptL.add(DigitalCash.NULL_KEY);
			encryptR.add(key);
		}
	}

	public void setAmount(int amt)
	{
		amount = amt;
	}

	public boolean isSpent()
	{
		return isSpent;
	}
	
	public void setSpent(boolean s)
	{
		isSpent = s;
	}

	// Returns the data of the money order in a string necessary for blinded signatures
	// Format:
	//	UniqueStr || Identity Strings || Amount
	public ByteArray intString()
	{
		byte [] dataStr = new byte[DigitalCash.MO_SIZE];
		byte [] tempData;
		int index = 0;

		// Copy the unique string
		tempData = uniqueStr.toByteArray();
		DigitalCash.CopyAllBytes(dataStr, tempData, 0);
		index += DigitalCash.UNIQUE_ENCRYPT_SIZE;

		// Copy the identity strings
		for (int i = 0; i < DigitalCash.N; i++) {
			DigitalCash.CopyAllBytes(dataStr, identityL.get(i).toByteArray(), index);
			index += DigitalCash.IDENTITY_ENCRYPT_SIZE;

			DigitalCash.CopyAllBytes(dataStr, identityR.get(i).toByteArray(), index);
			index += DigitalCash.IDENTITY_ENCRYPT_SIZE;
		}

		// Copy the amount
		tempData = DigitalCash.intToByteArray(amount);
		DigitalCash.CopyAllBytes(dataStr, tempData, index);

		// Append the hash
		return new ByteArray(dataStr);
	}

	public ByteArray hash()
	{
		return CryptoPrimitives.Hash(this.intString());
	}

	// Unsigns the money order with the bank's key and ensures it is valid
	public boolean isValidOrder()
	{
		if (signedHash == null) return false; // If there is no signed money order

		ByteArray unsigned = new ByteArray(CryptoPrimitives.RSAencrypt(signedHash.toInteger(), DigitalCash.getBankKey()));
		return CryptoPrimitives.CompareHash(unsigned, this.hash());
	}
}
