// Digital Cash
// Steven Kappes / Justin Fiedler

import java.math.*;
import java.security.*;
import java.security.interfaces.*;
import javax.crypto.*;

public class CryptoPrimitives
{
	private static final String RSA_MODE = "RSA/ECB/NoPadding";
	private static final String AES_MODE = "AES/ECB/NoPadding";
	private static final String HASH_MODE = "SHA-1";

	// Initialize a cipher
	private static Cipher getCipher(String algorithm, int mode, Key key)
	{
		try {
			Cipher cipher = Cipher.getInstance(algorithm);

			if (algorithm == AES_MODE) {
				// Initialize OFB mode
				AlgorithmParameters param = AlgorithmParameters.getInstance("AES");
				cipher.init(mode, key);
			} else {
				cipher.init(mode, key);
			}

			return cipher;
		} catch (Exception e) {
			System.err.println("Exception (getCipher): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}

	// Encrypt/decrypt with a cipher
	private static ByteArray runCipher(Cipher cipher, ByteArray in)
	{
		try {
			byte[] input = in.toByteArray();
			byte[] output;

			input = CryptoPrimitives.pad(cipher.getBlockSize(), input);
			output = cipher.doFinal(input, 0, input.length);

			return new ByteArray(output);
		} catch (Exception e) {
			System.err.println("Exception (runCipher): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}

	private static byte [] pad(int blockSize, byte [] in)
	{
		if (blockSize == 0) return in;
		if (in.length % blockSize == 0) return in;

		int addedBytes = (blockSize - (in.length % blockSize));
		byte [] padded = new byte[in.length + addedBytes];

		for (int i = 0; i < addedBytes; i++)
			padded[i] = (byte)0;

		DigitalCash.CopyAllBytes(padded, in, addedBytes);

		return padded;
	}

	public static SecretKey generateAESKey()
	{
		try {
			KeyGenerator gen = KeyGenerator.getInstance("AES");
			gen.init(128, new SecureRandom());
			return gen.generateKey();
		} catch (Exception e) {
			System.err.println("Exception (generateAESKey): " + e.getMessage());
			System.err.println(e.getStackTrace());
			System.exit(1);
		}
		return null;
	}

	// Encrypt with AES
	public static ByteArray AESencrypt(ByteArray in, Key key)
	{
		return CryptoPrimitives.runCipher(CryptoPrimitives.getCipher(AES_MODE, Cipher.ENCRYPT_MODE, key), in);
	}

	// Decrypt with AES
	public static ByteArray AESdecrypt(ByteArray in, Key key)
	{
		return CryptoPrimitives.runCipher(CryptoPrimitives.getCipher(AES_MODE, Cipher.DECRYPT_MODE, key), in);
	}

	// Initialize the hash function
	private static MessageDigest getHashFunction()
	{
		try {
			MessageDigest hash = MessageDigest.getInstance(HASH_MODE);		
			return hash;
		} catch (Exception e) {
			System.err.println("Exception (getHashFunction): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}

	// Returns the output size of the hash
	public static int GetHashSize()
	{
		try {
			MessageDigest hash = CryptoPrimitives.getHashFunction();		
			return hash.getDigestLength();
		} catch (Exception e) {
			System.err.println("Exception (GetHashSize): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return 0;
	}

	// Hash
	public static ByteArray Hash(ByteArray in)
	{
		try {
			MessageDigest hash = CryptoPrimitives.getHashFunction();
			byte [] data = hash.digest(in.toByteArray());
			data[0] = (byte)1; // WHYYYYYYY!??????????? is this necessary
			return new ByteArray(data);
		} catch (Exception e) {
			System.err.println("Exception (Hash): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}

	// Append the hash to the message
	public static ByteArray AppendHash(ByteArray in)
	{
		ByteArray hash = CryptoPrimitives.Hash(in);
		byte [] messageData = new byte[in.toByteArray().length + DigitalCash.HASH_SIZE];

		// Separate the message and hash value
		DigitalCash.CopyAllBytes(messageData, in.toByteArray(), 0);
		DigitalCash.CopyAllBytes(messageData, hash.toByteArray(), in.toByteArray().length);
		return new ByteArray(messageData);
	}

	// Compare hash
	public static boolean CompareHash(ByteArray in, ByteArray in2)
	{
		try {
			MessageDigest hash = CryptoPrimitives.getHashFunction();		
			return hash.isEqual(in.toByteArray(), in2.toByteArray());
		} catch (Exception e) {
			System.err.println("Exception (VerifyHash): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return false;		
	}

	// Return the message portion of a hash string
	public static ByteArray MessageOfHashString(ByteArray in)
	{
		byte [] messageData = new byte[in.toByteArray().length - DigitalCash.HASH_SIZE];
		byte [] hashData = new byte[DigitalCash.HASH_SIZE];

		// Separate the message and hash value
		DigitalCash.CopySomeBytes(messageData, in.toByteArray(), 0);
		DigitalCash.CopySomeBytes(hashData, in.toByteArray(), messageData.length);

		return new ByteArray(messageData);
	}

	// Return the hash portion of a hash string
	public static ByteArray HashOfHashString(ByteArray in)
	{
		byte [] messageData = new byte[in.toByteArray().length - DigitalCash.HASH_SIZE];
		byte [] hashData = new byte[DigitalCash.HASH_SIZE];

		// Separate the message and hash value
		DigitalCash.CopySomeBytes(messageData, in.toByteArray(), 0);
		DigitalCash.CopySomeBytes(hashData, in.toByteArray(), messageData.length);

		return new ByteArray(hashData);
	}

	// Verify the hash appended to a string
	public static boolean VerifyHashString(ByteArray in)
	{
		ByteArray message = CryptoPrimitives.MessageOfHashString(in);
		ByteArray hash = CryptoPrimitives.HashOfHashString(in);
		ByteArray hash2 = CryptoPrimitives.Hash(message);

		return CryptoPrimitives.CompareHash(hash, hash2);
	}

	// XOR two strings
	public static ByteArray XOR(ByteArray in, ByteArray in2)
	{
		return in.xor(in2);
	}

	// Generate a unique string of a certain length
	public static ByteArray UniqueString(int length)
	{
		byte [] bytes = new byte[length];

		for (int i = 0; i < length; i++)
			bytes[i] = (byte)(65.0 * Math.random() + 26.0); // Capital letters

		return new ByteArray(bytes);
	}

/////////////////////////////// RSA ///////////////////////////////////////////
// Encrypt with RSA
public static BigInteger RSAencrypt(BigInteger in, RSAPrivateKey key)
{
	//return CryptoPrimitives.runCipher(CryptoPrimitives.getCipher(RSA_MODE, Cipher.ENCRYPT_MODE, key), in);
	return in.modPow(key.getPrivateExponent(), key.getModulus());
}

public static BigInteger RSAencrypt(BigInteger in, RSAPublicKey key)
{
	//return CryptoPrimitives.runCipher(CryptoPrimitives.getCipher(RSA_MODE, Cipher.DECRYPT_MODE, key), in);
	return in.modPow(key.getPublicExponent(), key.getModulus());
}

public static ByteArray RSAencrypt(ByteArray in, ByteArray expon, ByteArray mod)
{	return null;//in.modPow(expon, mod);
}

/*
// Cipher decrypt
public static ByteArray RSAdecrypt(ByteArray in, RSAKey key)
{	return CryptoPrimitives.runCipher(CryptoPrimitives.getCipher(RSA_MODE, Cipher.DECRYPT_MODE, key), in);
}
*/

/////////////////////////////// BLINDING //////////////////////////////////////

	// Returns a random blinding factor 'r' 
	// GCD(r,n)=1 where 'n' is the modulus of the key
	public static BigInteger getNewBlindFactor(BigInteger n) {
		try { 
		SecureRandom random = SecureRandom.getInstance("SHA1PRNG", "SUN");


          byte [] randomBytes = new byte[10];
          BigInteger r = null;
          BigInteger gcd = null;
          BigInteger one = new BigInteger("1");

          //check that gcd(r,n) = 1 && r < n && r > 1
          do {
			random.nextBytes(randomBytes);
			r = new BigInteger(1, randomBytes);
			gcd = r.gcd(n);
			//System.out.println("gcd: " + gcd);
          }
          while(!gcd.equals(one) || r.compareTo(n)>=0 || r.compareTo(one)<=0);
		
		return r;
		} catch (Exception e) {
			System.err.println("Exception (getNewBlindFactor): " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}
	
	// RSA Blinds message 'm' with blinding factor 'r'
	public static BigInteger blind(BigInteger m, BigInteger r, RSAPublicKey pubkey) {
		BigInteger e = pubkey.getPublicExponent();
		BigInteger n = pubkey.getModulus();

		return ((r.modPow(e,n)).multiply(m)).mod(n);
	}

	// RSA Unblinds message 'bm' with blinding factor 'r'
	public static BigInteger unblind(BigInteger bm, BigInteger r, RSAPublicKey pubkey) {
		BigInteger n = pubkey.getModulus();

		return r.modInverse(n).multiply(bm).mod(n);
	}
}
