// Digital Cash
// Steven Kappes / Justin Fiedler

import java.io.*;
import java.util.*;
import java.net.*;
import java.math.*;
import javax.net.ssl.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.security.*;
import javax.crypto.*;
import org.w3c.dom.*;

// ***Messages***
// 0 = Balance
// 1 = Encrypted money orders
// 2 = Index which will be signed, unencrypt/unblind other orders
// 3 = Money order encryption keys and blinds
// 4 = Signed hash
// 5 = Money order to merchant
// 6 = Request identity keys
// 7 = Sends requested identity keys to merchant
// 8 = Deposit cash money order in bank

// 100 = Bad password
// 110 = Banned
// 120 = Generic message

public class XMLHelper
{	static private org.apache.xerces.dom.DOMImplementationImpl DOM = new org.apache.xerces.dom.DOMImplementationImpl();
	static private DocumentType docType = DOM.createDocumentType("message", "http://localhost/message.dtd", "message.dtd");

	// returns a new XML document
	public static Document newXMLDocument() {
		return newXMLDocument(docType);
	}

	// returns a new XML document of _docType
	public static Document newXMLDocument(DocumentType _docType) {
		return new org.apache.xerces.dom.DocumentImpl(_docType);
	}
	
	// constructs a new attribute node <tag attributeName=value>
	public static Element xmlAttributeNode(Document doc, String tag, String attributeName, String value){
		Element e = doc.createElement(tag);
		
		e.setAttribute(attributeName, value);

		return e;
	}

	//constructs a new <header msgNum=msgNum customerID=customerID>
	public static Element xmlHeader(Document doc, String msgNum, String customerID, String password){
		Element e = doc.createElement("header");
		
		e.appendChild(xmlAttributeNode(doc, "msgNum", "number", msgNum));
		if(customerID != null) e.appendChild(xmlAttributeNode(doc, "customerID", "id", customerID));
		if(password != null) e.appendChild(xmlAttributeNode(doc, "password", "text", password));

		return e;
	}

	public static Element xmlItem(Document doc, String tag, String index, String value){
		Element e = doc.createElement("item");
		
		e.setAttribute("tag", tag);
		e.setAttribute("index", index);
		e.setAttribute("value", value);
		
		return e;
	}

	//returns the value of <item tag=tag index=index value=?> if found
	public static String getItem(Node n, String tag, String index){
		String value = null;
		if(!n.hasChildNodes()){
			if(n.hasAttributes()){
				if(n.getNodeName().equals("item")){
					boolean flag = true;					
					NamedNodeMap att = n.getAttributes();
					for(int i = 0; i < att.getLength() && flag; i++){
						Node curAtt  = att.item(i);
						if(curAtt.getNodeName().equals("tag")){
							if(!curAtt.getNodeValue().equals(tag)) flag = false;
						} else if(curAtt.getNodeName().equals("index")) {
							if(!curAtt.getNodeValue().equals(index)) flag = false;
						} else if(curAtt.getNodeName().equals("value")) {
							value = curAtt.getNodeValue();
						}
					}
					if(flag == true) return value;
				}
			}
		} else {
			NodeList children = n.getChildNodes();
			for(int i = 0; i < children.getLength(); i++){
				value = getItem(children.item(i), tag, index);
				if(value != null) return value;
			}
		}
		return value;
	}

	// returns document header node
	public static Node getHeaderNode(Node doc){
		doc = doc.getLastChild();
		NodeList children = doc.getChildNodes();
		for(int i = 0; i < children.getLength(); i++){
			Node header = children.item(i);
			if(header.getNodeName().equals("header")){
				return header;
			}
		}
		return null;
	}

	// returns selected sub item of header
	public static Node getHeaderSubNode(Node header, String nodeName){
		Node subnode = null;
		NodeList children = header.getChildNodes();
		for(int i = 0; i < children.getLength(); i++){
			subnode = children.item(i);
			if(subnode.getNodeName().equals(nodeName)){
				return subnode;
			}
		}
		return subnode;
	}
	
	public static String getAttributeValue(Node n, String attribute){
		NamedNodeMap att = n.getAttributes();
		for(int i = 0; i < att.getLength(); i++){
			Node curAtt = att.item(i);
			if(curAtt.getNodeName().equals(attribute)) return curAtt.getNodeValue();
		}
		return null;
	}
	
	public static String getCustomerID(Document doc){
		return getHeaderItemValue(doc, "customerID", "id");
	}

	public static int getMessageNumber(Document doc){
		return Integer.parseInt(getHeaderItemValue(doc, "msgNum", "number"));
	}

	public static String getPassword(Document doc){
		return getHeaderItemValue(doc, "password", "text");
	}

	public static String getHeaderItemValue(Document doc, String item, String attribute){
		String value = null;
		Node header = getHeaderNode(doc);
		Node subNode = null;
		if(header != null) subNode = getHeaderSubNode(header, item);
		if(subNode != null) value = getAttributeValue(subNode, attribute);

		return value;
	}

	public static boolean isValidMessage(Document doc, int expectedMsg)
	{
		int msgNum = getMessageNumber(doc);

		if (msgNum == expectedMsg) {
			return true;
		} else if (msgNum == 100) {
			System.err.println("Unable to complete request: bad password");
		} else if (msgNum == 110) {
			System.err.println("Unable to complete request: you are banned");
		} else if (msgNum == 120) {
			System.err.println(getItem(doc, "Message", "0"));
		} else {
			System.err.println("Unable to process message " + msgNum + ": exptected " + expectedMsg);
		}
		return false;
	}

	// Message 0
	// Send/request the balance
	public static Document sendBalance(String customerID, String password, int bal)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "0", customerID, password));

		if (bal >= 0) message.appendChild(xmlItem(doc, "balance", "0", String.valueOf(bal)));

		return doc;
	}

	// Message 1
	// Customer sends encrypted money orders and blinded hashes
	public static Document sendEncryptedOrders(String customerID, String password,
				ArrayList<ByteArray> encryptedMOs, ArrayList<ByteArray> hashes)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "1", customerID, password));

		// Append the encrypted money orders
		for (int i = 0; i < encryptedMOs.size(); i++) {
			message.appendChild(xmlItem(doc, "EncryptedMoneyOrder", String.valueOf(i), encryptedMOs.get(i).toString()));
		}

		// Append the blinded hashes
		for (int i = 0; i < hashes.size(); i++) {
			message.appendChild(xmlItem(doc, "BlindHash", String.valueOf(i), hashes.get(i).toString()));
		}

		return doc;
	}
	
	// Message 2
	// Bank sends the index which will be signed
	public static Document sendMOIndex(int index)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "2", "d", "e"));
		message.appendChild(xmlItem(doc, "MOIndex", "0", String.valueOf(index)));
		return doc;
	}

	// Message 3
	// Customer sends list of keys and blinding factors
	public static Document sendKeysAndBlinds(String customerID, String password,
				ArrayList<SecretKey> keys, ArrayList<BigInteger> blinds,
				ArrayList<ArrayList<SecretKey>> leftKeys, ArrayList<ArrayList<SecretKey>> rightKeys)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "3", customerID, password));

		// Append the keys
		for (int i = 0; i < keys.size(); i++) {
			message.appendChild(xmlItem(doc, "MoneyOrderKey", String.valueOf(i), DigitalCash.getIntFromKey(keys.get(i)).toString()));
		}

		// Append the blinded hashes
		for (int i = 0; i < blinds.size(); i++) {
			message.appendChild(xmlItem(doc, "Blinds", String.valueOf(i), blinds.get(i).toString()));
		}
		
		// Apply the identity keys
		for (int i = 0; i < leftKeys.size(); i++) {
			for (int j = 0; j < leftKeys.get(i).size(); j++) { // Note: this is ridiculous
				message.appendChild(xmlItem(doc, "IdentKeyLeft" + String.valueOf(i), String.valueOf(j),
						 DigitalCash.getIntFromKey(leftKeys.get(i).get(j)).toString()));
				message.appendChild(xmlItem(doc, "IdentKeyRight" + String.valueOf(i), String.valueOf(j),
						 DigitalCash.getIntFromKey(rightKeys.get(i).get(j)).toString()));
			}
		}

		return doc;
	}

	// Message 4
	// Bank sends the signed hash to Customer
	public static Document sendSignedHash(ByteArray signedHash)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "4", null, null));
		message.appendChild(xmlItem(doc, "SignedHash", "0", signedHash.toString()));
		return doc;
	}

	// Message 5
	// Customer sends the MO and signed hash to Merchant
	public static Document sendMoneyAndSignedHash(ByteArray MO, ByteArray signedHash)
	{	
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "5", null, null));
		message.appendChild(xmlItem(doc, "MoneyOrder", "0", MO.toString()));
		message.appendChild(xmlItem(doc, "SignedHash", "0", signedHash.toString()));
		return doc;
	}

	// Message 6
	// Merchant requests key for one of each identity string pair
	public static Document requestIdentityKeys(ByteArray moUniStr, ArrayList<Integer> unlockSecs)
	{	Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "6", null, null));

		message.appendChild(xmlItem(doc, "MoneyOrderUniqueStr", "0", moUniStr.toString()));
		for(int i = 0; i < unlockSecs.size(); i++){
			message.appendChild(xmlItem(doc, "unlocksecret", String.valueOf(i), unlockSecs.get(i).toString()));
		}
		return doc;
	}

	// Message 7
	// Customer sends requested keys to Merchant
	public static Document sendIdentityKeys(ArrayList<SecretKey> identKeys)
	{	Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "7", null, null));
		for(int i = 0; i < identKeys.size(); i++){
			message.appendChild(xmlItem(doc, "IdentKey", String.valueOf(i), DigitalCash.getIntFromKey(identKeys.get(i)).toString()));
		}
		return doc;
	}

	// Message 8
	// Merchant sends money order to Bank
	public static Document cashMoneyOrder(String merchantID, String password, ByteArray MO, ByteArray signedHash,
							 ArrayList<Integer> unlockSecs, ArrayList<SecretKey> identKeys)
	{	Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "8", merchantID, password));
		message.appendChild(xmlItem(doc, "MoneyOrder", "0", MO.toString()));
		message.appendChild(xmlItem(doc, "SignedHash", "0", signedHash.toString()));
		for(int i = 0; i < unlockSecs.size(); i++){
			message.appendChild(xmlItem(doc, "unlocksecret", String.valueOf(i), unlockSecs.get(i).toString()));
		}
		for(int i = 0; i < identKeys.size(); i++){
			message.appendChild(xmlItem(doc, "IdentKey", String.valueOf(i), DigitalCash.getIntFromKey(identKeys.get(i)).toString()));
		}
		return doc;
	}

	// Message 100
	public static Document badPassword()
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "100", null, null));
		return doc;
	}

	// Message 110
	public static Document banned()
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "110", null, null));
		return doc;
	}

	// Message 120
	public static Document sendMessage(String msg)
	{
		Document doc = newXMLDocument();
		Element message = doc.createElement("message");
		doc.appendChild(message);
		message.appendChild(xmlHeader(doc, "120", null, null));
		message.appendChild(xmlItem(doc, "Message", "0", msg));
		return doc;
	}

	//recursively prints XML document tree
	public static void printXMLTree(Node n){
		if(!n.hasChildNodes()){
			if(n.hasAttributes()){
				NamedNodeMap att = n.getAttributes();
				for(int i = 0; i < att.getLength(); i++){
					System.out.println(att.item(i).getNodeName() +": "+ att.item(i).getNodeValue());
				}
			}
		} else {
			System.out.println("Node: " + n.getNodeName());
			NodeList children = n.getChildNodes();
			for(int i = 0; i < children.getLength(); i++){
				printXMLTree(children.item(i));
			}
		}
	}

	//removes empty text nodes
    public static void stripText(Node n)
	{
		NodeList l = n.getChildNodes();

		for (int i = 0; i<l.getLength(); ++i) {
			Node nn = l.item(i);
			if (nn.getNodeType() == nn.TEXT_NODE) {
				n.removeChild(nn);
			}
	    }
	}
}
