// Digital Cash
// Steven Kappes / Justin Fiedler

import java.util.*;
import java.math.*;
import java.security.*;
import javax.crypto.*;
import java.security.interfaces.*;

public class Bank
{
	// Subclass structure to store widthdrawals
	private class WithdrawalRequest
	{
		public ByteArray customerID;
		public ArrayList<ByteArray> encryptedMOs;
		public ArrayList<ByteArray> blindedHashes;
		public ArrayList<MoneyOrder> requestedOrders;
		public ArrayList<ByteArray> requestedMOHashes;
		public boolean readyToVerify;
		public int futureMOIndex; // Index of the money order which will be signed

		public WithdrawalRequest()
		{
			readyToVerify = false;
			futureMOIndex = -1;
			encryptedMOs = new ArrayList<ByteArray>();
			blindedHashes = new ArrayList<ByteArray>();
			requestedOrders = new ArrayList<MoneyOrder>();
			requestedMOHashes = new ArrayList<ByteArray>();
		}
	}

	// Account information
	private Hashtable<String, Integer> balances;

	// Bank state
	private ArrayList<ByteArray> bannedAccounts;
	private ArrayList<WithdrawalRequest> withdrawals;
	private ArrayList<MoneyOrder> deposits;

	public String ERROR = "";

	public Bank()
	{
		balances = new Hashtable<String, Integer>();
		withdrawals = new ArrayList<WithdrawalRequest>();
		deposits = new ArrayList<MoneyOrder>();
		bannedAccounts = new ArrayList<ByteArray>();
	}

	public int receivedMoneyOrders(ByteArray custID, ArrayList<ByteArray> moneyOrders, ArrayList<ByteArray> hashes)
	{
		WithdrawalRequest wdr = new WithdrawalRequest();
		ArrayList<ByteArray> encryptedMOs = new ArrayList<ByteArray>();
		ArrayList<ByteArray> blindedHashes = new ArrayList<ByteArray>();

		if (this.getRequestForCustomer(custID) != null) {
			System.err.println("Bank.receivedMoneyOrders: Money order request already pending");
			return -1;
		}

		if (this.isBanned(custID)) {
			ERROR = "Unable to receive money orders: " + (new String(custID.toByteArray())) + " is banned";
			return -1;
		}

		for (int i = 0; i < moneyOrders.size(); i++) {
			// Save the encrypted money orders
			encryptedMOs.add(moneyOrders.get(i));

			// Save the hashes
			blindedHashes.add(hashes.get(i));
		}

		// Save the withdrawal request
		wdr.customerID = new ByteArray(custID.toByteArray());
		wdr.encryptedMOs = encryptedMOs;
		wdr.blindedHashes = blindedHashes;
		wdr.futureMOIndex = (int)(Math.random() * moneyOrders.size());	
		if (wdr.futureMOIndex >= moneyOrders.size()) {
			System.err.println("Bank.receivedMoneyOrders: Invalid money order to unblind");
			return -1;
		}
		withdrawals.add(wdr);

		// Request unblind
		return wdr.futureMOIndex;
	}
	
	public boolean receivedEncryptionKeys(ByteArray custID, ArrayList<SecretKey> keys)
	{
		WithdrawalRequest wdr = this.getRequestForCustomer(custID);

		if (wdr == null) {
			ERROR = "No money order pending";
			return false;
		}

		for (int i = 0; i < wdr.encryptedMOs.size(); i++) {
			if (i != wdr.futureMOIndex) {
				int decryptIndex = i;
				ByteArray decryptedMO;
				MoneyOrder mo;

				if (i > wdr.futureMOIndex) decryptIndex--; // The unblind factors will be offset by 1

				// Decrypt 
				decryptedMO = CryptoPrimitives.AESdecrypt(wdr.encryptedMOs.get(i), keys.get(decryptIndex));

				if (decryptedMO == null) { // Decryption failed
					withdrawals.remove(wdr);
					return false;
				} else {
					mo = new MoneyOrder(null, decryptedMO);
					wdr.requestedOrders.add(mo); 
				}
			} else {
				wdr.requestedOrders.add(new MoneyOrder()); // Add a placeholder money order
			}
		}

		return true;
	}

	// Unblind the requested money orders hashes
	public boolean receivedUnblindFactors(ByteArray custID, ArrayList<BigInteger> factors)
	{
		WithdrawalRequest wdr = this.getRequestForCustomer(custID);
		RSAPublicKey myKey = DigitalCash.getBankKey();

		if (wdr == null) {
			ERROR = "No money order pending";
			return false;
		}

		for (int i = 0; i < wdr.encryptedMOs.size(); i++) {
			if (i != wdr.futureMOIndex) {
				int unblindIndex = i;
				BigInteger unblinded;
				ByteArray unblindedHash;
				MoneyOrder mo;

				if (i > wdr.futureMOIndex) unblindIndex--; // The unblind factors will be offset by 1

				// Unblind
				unblinded = new BigInteger(1, wdr.blindedHashes.get(i).toByteArray()); // Get the hash
				unblinded = CryptoPrimitives.unblind(unblinded,
					factors.get(unblindIndex).modPow(myKey.getPublicExponent(), myKey.getModulus()), myKey);
				unblindedHash = new ByteArray(unblinded.toByteArray());

				if (unblindedHash == null) { // Unblinding failed
					withdrawals.remove(wdr);
					ERROR = "Failed to unblind hashes";
					return false;
				} else {
					wdr.requestedMOHashes.add(unblindedHash);
				}
			} else {
				wdr.requestedMOHashes.add(new ByteArray()); // Add a placeholder
			}
		}

		wdr.readyToVerify = true;
		return true;
	}

	public boolean receivedKeys(ByteArray custID, int moIndex, ArrayList<SecretKey> encryptL, ArrayList<SecretKey> encryptR)
	{
		// Save the encryption keys for the money order index
		WithdrawalRequest wdr = this.getRequestForCustomer(custID);
		MoneyOrder mo;

		if (wdr == null) {
			ERROR = "No money order pending";
			return false;
		}

		if (!wdr.readyToVerify) {
			ERROR = "Required information for the money order missing";
			return false;
		}

		try {
			mo = wdr.requestedOrders.get(moIndex);
			ArrayList<SecretKey> enL = mo.encryptKeysLeft();			
			ArrayList<SecretKey> enR = mo.encryptKeysRight();

			for (int i = 0; i < DigitalCash.N; i++) {
				enL.add(encryptL.get(i));
				enR.add(encryptR.get(i));
			}
		} catch (Exception e) {
			System.err.println("Bank.receivedKeys: Error saving keys");
			System.exit(1);
		}
		return true;
	}

	public ByteArray getSignedMoneyOrder(ByteArray custID)
	{
		// Save the encryption keys for the money order index
		WithdrawalRequest wdr = this.getRequestForCustomer(custID);
		boolean success = true;
		int moAmount = -1;

		if (wdr == null) {
			ERROR = "No money order pending";
			return null;
		}

		if (!wdr.readyToVerify) {
			ERROR = "Required information for the money order missing";
			return null;
		}

		try {
			// Verify the money orders are valid
			for (int i = 0; ((i < DigitalCash.N) && (success)); i++) {
				if (i != wdr.futureMOIndex) {
					MoneyOrder mo = wdr.requestedOrders.get(i);
					ByteArray leftIdent, rightIdent, finalIdent;

					if (moAmount == -1) {
						moAmount = mo.amount();
					}

					// Verify the amount
					if (moAmount != mo.amount())
						success = false;

					// Verify the identity strings & encryption keys
					for (int j = 0; ((j < DigitalCash.N) && (success)); j++) {
						leftIdent = mo.identityStringsLeft().get(j);
						rightIdent = mo.identityStringsRight().get(j);

						// Decrypt
						leftIdent = CryptoPrimitives.AESdecrypt(leftIdent, mo.encryptKeysLeft().get(j));
						rightIdent = CryptoPrimitives.AESdecrypt(rightIdent, mo.encryptKeysRight().get(j));

						// Verify the left hash
						if (!CryptoPrimitives.VerifyHashString(leftIdent))
							success = false;

						// Verify the right hash
						if (!CryptoPrimitives.VerifyHashString(rightIdent))
							success = false;

						// Verify the identity strings
						finalIdent = CryptoPrimitives.XOR(CryptoPrimitives.MessageOfHashString(leftIdent),
								CryptoPrimitives.MessageOfHashString(rightIdent));

						if (!finalIdent.equals(custID))
							success = false;
					}
				}
			}

			// Verify the hashes are valid
			for (int i = 0; ((i < DigitalCash.N) && (success)); i++) {
				if (i != wdr.futureMOIndex) {
					MoneyOrder mo = wdr.requestedOrders.get(i);
					ByteArray hash = mo.hash();
					ByteArray sentHash = wdr.requestedMOHashes.get(i);

					// Verify the hash the customer sent is correct
					if (!CryptoPrimitives.CompareHash(hash, sentHash))
						success = false;
				}
			}
		} catch (Exception e) {
			ERROR = "Verification of money order failed";
			return null;
		}

		// Remove the withdrawal request
		withdrawals.remove(wdr);

		if (success) {
			ByteArray blindedMOHash = wdr.blindedHashes.get(wdr.futureMOIndex);

			// Withdrawal the money from the account
			if (!this.withdraw(custID, moAmount)) {
				ERROR = "The balance has been overdrawn";
				return null; // Balance not large enough
			}

			// Sign the remaining money order and return it
			return new ByteArray(CryptoPrimitives.RSAencrypt(blindedMOHash.toInteger(), DigitalCash.getMyPrivateKey()));
		} else {
			// Banned
			this.banAccount(custID);
			ERROR = "Invalid money orders: you are banned";
		}
		return null;
	}

	public boolean depositMoneyOrder(ByteArray merchID, MoneyOrder mo)
	{
		MoneyOrder dep = this.getDeposit(mo);

		if (!mo.isValidOrder()) {
			this.banAccount(merchID);
			ERROR = "Signature not valid: you are banned";
			return false;
		}
		if (DigitalCash.N != mo.unlockedIdentitySecrets().size()) return false;

		if (dep == null) {
			// Check the provided keys are correct
			for (int i = 0; i < DigitalCash.N; i++) {
				if (!this.verifyIdentKey(mo, i, (mo.unlockedIdentitySecrets().get(i).intValue() == 0))) { // bad merchant
					ERROR = "Identity string verification failed";
					this.banAccount(merchID);
					return false;
				}
			}

			// Money order not yet deposited, the sent keys are valid, accept the money order
			deposits.add(mo);
			this.withdraw(merchID, -mo.amount());
			return true;
		}

		// Determine who is at fault, check the unlocked identities
		// dep is the previously deposited order
		// mo is the current deposit
		// check to see if the identity 
		for (int i = 0; i < DigitalCash.N; i++) {
			int b1 = mo.unlockedIdentitySecrets().get(i).intValue();
			int b2 = dep.unlockedIdentitySecrets().get(i).intValue();

			if (b1 != b2) { // Customer is using the same money order twice
				SecretKey left = ((b1 == 0) ? mo.encryptKeysLeft().get(i) : dep.encryptKeysLeft().get(i));
				SecretKey right = ((b2 == 0) ? mo.encryptKeysRight().get(i) : dep.encryptKeysRight().get(i));

				ByteArray sec1 = dep.identityStringsLeft().get(i);
				ByteArray sec2 = dep.identityStringsRight().get(i);
				ByteArray ident;

				// Unencrypt the secrets
				sec1 = CryptoPrimitives.AESdecrypt(sec1, left);
				sec2 = CryptoPrimitives.AESdecrypt(sec2, right);

				if ((!CryptoPrimitives.VerifyHashString(sec1)) || (!CryptoPrimitives.VerifyHashString(sec2))) {
					// The merchant sent bad keys
					ERROR = "Identity string verification failed: you are banned";
					this.banAccount(merchID);
					return false;
				}

				sec1 = CryptoPrimitives.MessageOfHashString(sec1); // Get only the secret
				sec2 = CryptoPrimitives.MessageOfHashString(sec2);

				ident = CryptoPrimitives.XOR(sec1, sec2); // Customer identity has been established
				ERROR = "Unable to deposit: duplicate money order by customer (" + (new String(ident.toByteArray())) + ")";
				this.banAccount(ident); // Goodbye
				return false;
			}
		}

		// Merchant is using the same money order twice
		ERROR = "Unable to deposit: duplicate money order by merchant (" + (new String(merchID.toByteArray())) + ")";
		this.banAccount(merchID);
		return false;
	}

	public int pendingMoneyOrder(ByteArray custID)
	{
		WithdrawalRequest wdr = this.getRequestForCustomer(custID);

		return (wdr == null ? -1 : wdr.futureMOIndex);
	}

	private WithdrawalRequest getRequestForCustomer(ByteArray custID)
	{
		for (int i = 0; i < withdrawals.size(); i++) {
			if (withdrawals.get(i).customerID.equals(custID)) {
				return withdrawals.get(i);
			}
		}
		return null; // Request not found
	}

	private MoneyOrder getDeposit(MoneyOrder mo)
	{
		for (int i = 0; i < deposits.size(); i++) {
			if (mo.uniqueString().equals(deposits.get(i).uniqueString())) return deposits.get(i);
		}

		return null;
	}

	private void banAccount(ByteArray custID)
	{
		if (!this.isBanned(custID)) {
			bannedAccounts.add(custID);
			System.out.println("BANNED: " + new String(custID.toByteArray()));
		}
	}

	public boolean isBanned(ByteArray custID)
	{
		for (int i = 0; i < bannedAccounts.size(); i++)
			if (bannedAccounts.get(i).equals(custID)) return true;

		return false;
	}

	private int createBalance(ByteArray custID)
	{
		int amount = 10000; // $10,000 initially
		try {
			balances.put(custID.toString(), new Integer(amount));
			return amount;
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		return 0;
	}

	public int getBalance(ByteArray custID)
	{
		try {
			Integer amt = balances.get(custID.toString());
			if (amt == null) return this.createBalance(custID);
			return amt.intValue();
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		return 0;
	}

	private boolean withdraw(ByteArray custID, int amount)
	{
		try {
			int curAmount = this.getBalance(custID);

			// Account is overdrawn
			if (curAmount < amount) return false;

			// Withdraw the money
			balances.remove(custID.toString());
			balances.put(custID.toString(), new Integer(curAmount - amount));
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		return false;
	}

	private boolean verifyIdentKey(MoneyOrder mo, int index, boolean left)
	{
		if ((index < 0) || (index >= DigitalCash.N)) return false;

		SecretKey key = ((left) ? mo.encryptKeysLeft().get(index) : mo.encryptKeysRight().get(index));
		ByteArray sec = ((left) ? mo.identityStringsLeft().get(index) : mo.identityStringsRight().get(index));

		// Unencrypt the secrets
		sec = CryptoPrimitives.AESdecrypt(sec, key);

		return CryptoPrimitives.VerifyHashString(sec);
	}
}
